
asset_list = {}


class asset_enroller(object):
    def __init__(self, cls, *args, **kwargs):
        asset_list.update({
            cls.__name__: cls
        })

        self.cls = cls

    def __call__(self, *args, **kwargs):
        pass



"""Populate the property objects
"""
import copy
import json
from abc import ABCMeta, abstractmethod

from compipe.exception.validate_error import GErrorNullObject, GErrorValue, GErrorKeyNotFound
from compipe.response.command_result import MSGStatusCodes
from compipe.utils.logging import logger
from compipe.utils.parameters import (ARG_DATA, ARG_FILE, ARG_GUID,
                                      ARG_NAME, ARG_OBJ, ARG_PARENT)
from compipe.utils.task_queue_helper import TQHelper
from pydantic.error_wrappers import ValidationError

from ..github_app.github_helper import JsonPropertiesHelper
from ..hash_code_helper import hexdigest_str
from .resolver_parameter import ResolverParam

IDENTITY_KEYS = {'id', 'name'}


class AbstractAssetResolver(metaclass=ABCMeta):

    @abstractmethod
    def populate(self):
        """Populate asset to full version config.
        """
        pass

    @abstractmethod
    def validate(self):
        """Validate content when populating
        """
        pass


class AssetResolver(AbstractAssetResolver):
    def __init__(self, param: ResolverParam):
        self._repo_helper = None
        self.ignore_paths = param.ignore_paths
        self.main_branch = param.main_branch
        self.param = param

        # validate schema-json
        # e.g. validate(instance=value, schema=self.schema)
        self.model = param.model
        self.schema = param.model.schema_json(indent=4)
        self.configs = {}
        self.dependency = {}

        # it's aimed to cache the resolved config when building dependencies
        self._cached_configs = {}

    @property
    def repo_helper(self):
        if not self._repo_helper:
            self._repo_helper = JsonPropertiesHelper(self.param.repo,
                                                     self.param.filter_source,
                                                     self.param.filter_export,
                                                     self.param.output_path,
                                                     self.param.base_url)
        return self._repo_helper

    def get(self, path):
        return self.configs.get(path)

    def load_configs(self):
        self.configs = self.repo_helper.get_properties(branch=self.main_branch,
                                                       filters=self.param.filter_source)

        for cfg in [cfg for _, cfg in self.configs.items() if not cfg.get(ARG_PARENT)]:
            obj = self.model.parse_obj(cfg.get(ARG_DATA))
            cfg[ARG_DATA] = json.loads(obj.json())

    def populate(self):

        self.load_configs()

        # start to build inheritance dependencies
        for path, value in self.configs.items():
            branches = self.build_dependencies(path, value, [])
            route = '.'.join(branches)
            if route not in self.dependency:
                self.dependency.update({
                    route: branches
                })

        # start to build all configs
        # the building routes would be:
        # root -> parent -> child
        #                -> child
        #                -> child -> sub-child
        #                         -> sub-child
        for routes in self.dependency.values():
            for path in routes:

                if path in self._cached_configs:
                    continue

                config = self.configs.get(path)
                if config.get(ARG_PARENT):
                    parent_config = self.configs.get(config[ARG_PARENT])
                    populated_config_data = copy.deepcopy(parent_config.get(ARG_DATA))

                    # start to override the sub-config properties
                    self.override_data(config[ARG_DATA], populated_config_data)
                    populated_config_data[ARG_GUID] = self._gen_guid(path)

                else:
                    populated_config_data = copy.deepcopy(config.get(ARG_DATA))

                asset = self.validate(path, populated_config_data)
                self.configs[path][ARG_DATA] = dict(populated_config_data)
                self.configs[path][ARG_OBJ] = asset

                # mark the built config as cached
                self._cached_configs[path] = True

    def _gen_guid(self, data):
        return next(iter(hexdigest_str(data)), None)

    @classmethod
    def override_data(cls, source: dict, target: dict):
        """Perform the resolver on dict object. The child object inherits values
        from parent

        Arguments:
            source {dict} -- [description]
            target {dict} -- [description]
        """
        for key, value in source.items():
            # process the dictionary type object
            if isinstance(value, dict):
                if key in target:

                    if not target[key]:
                        target[key] = value
                    else:
                        cls.override_data(value, target[key])
                        continue
            # process the array type object
            elif isinstance(value, list) and key in target:
                if not str(key).startswith('uid_'):
                    value = cls.resolve_list_items(value, target[key])

            target.update({
                key: value
            })

    @classmethod
    def resolve_list_items(cls, source, target):
        """Perform the resolver on 'list' property. 

        Comparing rules:
            Dict type: The dict item would be using "name" key as an 
            identifer to compare the values between child and par

        Arguments:
            source {list} -- Represent the source list.
            target {list} -- Represent the target list.

        Returns:
            list -- Represent the resolved new lists.
        """
        new_list = []
        # check element type
        list_item = next(iter(source + target), None)
        if isinstance(list_item, dict):
            # copy the items from source that don't exist in target
            for item in source:
                if not any(item[ARG_NAME] == t_item[ARG_NAME] for t_item in target):
                    new_list.append(item)
            # resolve the target items
            for item in target:
                # parse the valid identity key to compare the list elements
                if (identity_key := next(iter(IDENTITY_KEYS.intersection(item.keys())), None)) is None:
                    raise GErrorKeyNotFound(f'Not found one of the valid identity keys from {IDENTITY_KEYS}')

                source_item = next(iter([src for src in source if src[identity_key] == item[identity_key]]), None)
                if source_item:
                    # copy the source list item to the target
                    cls.override_data(source=source_item, target=item)

                # add the resolved target item to the new lists
                new_list.append(item)

        else:
            new_list = list(set(source + target))

        # # sort 'str' lists
        # if len(new_list) > 0 and isinstance(new_list[0], str):
        #     new_list.sort()

        return new_list

    def build_dependencies(self, path, config, branches=[]):
        """Build asset hierarchy dependencies. It would describe the relationship
        between children and parent object.

        Arguments:
            path {str} -- Represent the object path.
            config {dict} -- Represent the object value.

        Keyword Arguments:
            branches {list} -- Represent the route paths (default: {[]})

        Returns:
            list -- Represent the route dependencies.
        """
        logger.debug(f'Build dependencies on: [{path}]')
        parent = config.get(ARG_PARENT)
        branches.insert(0, path)
        if parent:
            parent_config = self.configs.get(parent)
            if not parent_config:
                raise GErrorNullObject(
                    f'Found invalid parent:[{self.configs.get(ARG_FILE)}] -> [{parent}]')
            return self.build_dependencies(parent, parent_config, branches)
        else:
            return branches

    def validate(self, path, dataset):
        try:
            # for path, value in self.configs.items():
            # ignore the "template" or some other customized ignored configs
            if not any(map(lambda x: path.startswith(x), self.ignore_paths)):

                # check parent object
                parent = dataset.get(ARG_PARENT)
                if parent and parent not in self.configs:
                    raise GErrorValue(f'Parent path is invalid: [{parent}]')

                # resolve object from metadata json
                return self.model.parse_obj(dataset)

        except ValidationError as e:
            TQHelper.post(message=f'Failed to validate object [{path}]',
                          payload=str(e),
                          msg_status=MSGStatusCodes.error)
            raise e



"""Publish properties (configs) from github to google drive
"""

import os

from compipe.exception.general_warning import WarningGotEmptyValue
from compipe.exception.validate_error import GErrorKeyNotFound
from compipe.response.command_result import MSGStatusCodes
from compipe.utils.io_helper import json_loader
from compipe.utils.logging import logger
from compipe.utils.parameters import ARG_OBJ, ARG_OUTPUT, ARG_DATA
from compipe.utils.singleton import Singleton
from compipe.utils.task_queue_helper import TQHelper
from typing import Dict

from .resolver.asset_resolver import AssetResolver, ResolverParam
from .resolver.resolver_parameter import ResolverParam

"""

Below config is to declare the properties repos.
It could help tars service to address the incoming webhook event
The incoming webhook URL should be ending with one of the key name of below declarations

ignore_path: it would skip building the configs when matching to the ignore lists.

"""

# collect the full lists of model names, it would be used to
# verify the class name parameter
# /////////////////////////////////////////////////////////////////////


class AssetPopulator(metaclass=Singleton):

    def __init__(self, resolver_cfg_path: str):
        self.resolver_cfg_path = resolver_cfg_path
        self.resolver_cfg_data = json_loader(self.resolver_cfg_path)
        self.resolvers = {key: ResolverParam.parse_obj(value) for key, value in self.resolver_cfg_data.items()}

    def get_resolver_param(self, model_name: str) -> ResolverParam:
        if not (resolver_param_data := self.resolver_cfg_data.get(model_name, None)):
            raise GErrorKeyNotFound(f'Not found model: {model_name}')

        return ResolverParam.parse_obj(resolver_param_data)

    def get_resolver(self, model_name: str):
        self.get_resolver_param(model_name=model_name)

    def clean_up_temporary_keys(self, data: Dict):

        # tmp attribute name has '_' prefix
        _tmp_keys = [key for key in data.keys() if key[0] == '_']

        # remove keys
        for _key in _tmp_keys:
            # remove temporary key from the data config
            del data[_key]

        return data

    def populate(self, resolver: AssetResolver, filters: list):
        """Generate the published configs to github.

        Arguments:
            filters {list} -- Represent the config lists.

        Returns:
            [list] -- Represent the populated config lists, '[]' empty lists
            means 
        """

        # keep all the influenced configs that will be exported later
        # //////////////////////////////////////////////////////////////////
        cfg_stacks = set()

        if not filters:
            # build all configs on the specific resolver
            # //////////////////////////////////////////////////////////////
            cfg_stacks.update([item for sublist in resolver.dependency.values() for item in sublist])
        else:
            # only build the selected configs of the resolver
            # //////////////////////////////////////////////////////////////
            for filter_cfg in filters:
                for _, configs in resolver.dependency.items():

                    # the filter item is represented by the config basename,
                    # we would need to re-format the lists before compareing
                    # //////////////////////////////////////////////////////
                    shorten_cfg_lists = [os.path.split(cfg)[-1] for cfg in configs]
                    if filter_cfg in shorten_cfg_lists:
                        index = shorten_cfg_lists.index(filter_cfg)
                        cfg_stacks.update(configs[index:])

        if not cfg_stacks:
            raise WarningGotEmptyValue('No config need to be resolved!')

        for path, config in resolver.configs.items():
            if path in cfg_stacks:
                logger.debug(f'Populate object: [{path}]')
                asset = config.get(ARG_OBJ)

                if asset and asset.publish > 0:
                    # if the file does exist, it would load the sha hash value for
                    # perform 'update' instead of 'add' on git.
                    # /////////////////////////////////////////////////////////////////////
                    sha = resolver.repo_helper.sha_mappings.get(config[ARG_OUTPUT])

                    # clean up data before committing
                    config_data = self.clean_up_temporary_keys(data=config[ARG_DATA])

                    # commit the changes to git repo
                    # /////////////////////////////////////////////////////////////////////
                    (is_uploaded, msg) = resolver.repo_helper.commit(output=config[ARG_OUTPUT],
                                                                     data=config_data,
                                                                     sha=sha,
                                                                     branch=resolver.main_branch)

                    TQHelper.post(payload=msg,
                                  msg_status=MSGStatusCodes.success if is_uploaded else MSGStatusCodes.warning)

                else:
                    # ignore unpublished asset
                    # /////////////////////////////////////////////////////////////////////
                    TQHelper.post(payload=f'Skipped unpublished config: {path}',
                                  msg_status=MSGStatusCodes.warning)

        return 'Done'



from pydantic import BaseModel, Field


class ClipRangeInt(BaseModel):
    """[Deprecated: use 'Threshold' to replace] Represent the threshold 
    Values (Region of interest).

    """
    start: int = Field(...,
                       title='The lower threshold of Clip Range',
                       ge=0,
                       le=10000)
    end: int = Field(...,
                     title='The upper threshold of Clip Range',
                     ge=0,
                     le=10000)

    class Config:
        schema_extra = {
            'examples': [
                {
                    'start': 0,
                    'end': 100
                }
            ]
        }


class ThresholdUnsignedRange(BaseModel):
    """Represent the threshold range values (Region of interest)

    """
    min: int = Field(...,
                     title='The lower threshold of the Range',
                     ge=0,
                     le=10000)
    max: int = Field(...,
                     title='The upper threshold of the Range',
                     ge=0,
                     le=10000)

    @property
    def values(self):
        return (self.min, self.max)

    @property
    def is_valid(self):
        return self.max > self.min

    class Config:
        schema_extra = {
            'examples': [
                {
                    'min': 0,
                    'max': 3000
                }
            ]
        }


class ThresholdRange(BaseModel):
    """Represent the threshold range values (Region of interest)

    """
    min: int = Field(...,
                     title='The lower threshold of the Range',
                     ge=-10000,
                     le=10000)
    max: int = Field(...,
                     title='The upper threshold of the Range',
                     ge=-10000,
                     le=10000)

    @property
    def values(self):
        return (self.min, self.max)

    @property
    def is_valid(self):
        return self.max > self.min

    def set_values(self, range=(0, 0)):
        self.min = range[0]
        self.max = range[1]

    class Config:
        schema_extra = {
            'examples': [
                {
                    'min': 0,
                    'max': 3000
                }
            ]
        }


class VolumeFloatRange(BaseModel):
    """Represent the volume range values (Unit: M3)

    """
    min: float = Field(...,
                       title='The lower threshold of the Range',
                       ge=0,
                       le=1)
    max: float = Field(...,
                       title='The upper threshold of the Range',
                       ge=0,
                       le=1)

    @property
    def values(self):
        return (self.min, self.max)

    @property
    def is_valid(self):
        return self.max > self.min

    def set_values(self, range=(0, 0)):
        self.min = range[0]
        self.max = range[1]

    class Config:
        schema_extra = {
            'examples': [
                {
                    'min': 0.0001,
                    'max': 1
                }
            ]
        }


class ThresholdRange3D(BaseModel):
    """Represent the threshold range values (Region of interest)

    Use cases:
        Medical Dicom: Define the borders of clipping 3d volume data (dicom dataset)
           width: Represent the width of axial plane.
           height: Represent the height of axial plane.
           depth: Represent the height of coronal/sagittal plane.
           offset_x: Represent the border offset value along axial plane.
           offset_y: Represent the border offset value along axial plane.
           offset_z: Represent the border offset value along coronal/sagittal plane

    """
    width: int = Field(...,
                       title='Represent the width along the X-Axis',
                       ge=0,
                       le=2000)
    height: int = Field(...,
                        title='Represent the height along the Y-Axis',
                        ge=0,
                        le=2000)

    depth: int = Field(...,
                       title='Represent the depth along the Z-Axis',
                       ge=0,
                       le=2000)

    offset_x: int = Field(...,
                          title='Represent the border offset value along X-Axis',
                          ge=0,
                          le=2000)
    offset_y: int = Field(...,
                          title='Represent the border offset value along Y-Axis',
                          ge=0,
                          le=2000)

    offset_z: int = Field(...,
                          title='Represent the border offset value along Z-Axis',
                          ge=0,
                          le=2000)

    class Config:
        schema_extra = {
            'examples': [
                {
                    'width': 0,
                    'height': 0,
                    'depth': 0,
                    'offset_x': 0,
                    'offset_y': 0,
                    'offset_z': 0
                }
            ]
        }

    @property
    def is_valid(self):
        return (self.width and self.height and self.depth)

    @property
    def values(self):
        return ((self.width, self.height, self.depth),
                (self.offset_x, self.offset_y, self.offset_z))

    def __repr__(self):
        return str(self.values)



import hashlib
import base64
import json
from cryptography.fernet import Fernet


def encrypt_str(value):
    key = Fernet.generate_key()
    cipher_suite = Fernet(key)
    cipher_text = cipher_suite.encrypt(bytes(value, 'utf-8'))
    return {'FERNET_KEY': key.decode("utf-8"),
            'CIPHER_TEXT': cipher_text.decode("utf-8")}


def decrypt_str(key, cipher_text):
    if not key or not cipher_text:
        return None
    else:
        cipher_suite = Fernet(bytes(key, 'utf-8'))
        return json.loads(cipher_suite.decrypt(bytes(cipher_text, 'utf-8')))


def md5_string_to_hash(*args, **kwargs):
    """Simply convert string to md5 hash value.

    Args:
        args (list): Represent the string list for converting
        kwargs (Dict): key: 'encode' represent encoding standard.
                        Default value: utf-8
    Example:
        >>>tars md5hash -a "do less shit, do more awesome shit" encode=utf-8
    """
    encoding = kwargs['encode'] if hasattr(kwargs, 'encode') else 'utf-8'
    return [hashlib.md5(item.encode(encoding)).hexdigest() for item in args]


def hexdigest_str(*args):

    return [hashlib.sha224(bytes(item, 'utf-8')).hexdigest() for item in args]


def convert_string_to_base64(*args):
    """Simply convert string to base64.

    Example:
        tars str2b64 -a "hello world"
    """

    string = base64.b64encode(bytes(args[0], 'utf-8'))
    return string.decode("utf-8")



import hashlib
import hmac
import json
import os
import pickle
import re
import shlex
from collections import defaultdict

from compipe.cmd_wrapper import create_command_payload
from compipe.utils.decorators import (required_params,
                                      verify_webhook_secret_sha1)
from compipe.utils.decorators_exception_handler import exception_handler
from compipe.utils.parameters import ARG_CLASS, ARG_PAYLOAD, ARG_POPULATE


class Generic:
    """Resolve the json string into a generic class instance
    It could allow to use class property syntax to access the key/value from dict.
    """
    @classmethod
    def from_dict(cls, dict):
        obj = cls()
        obj.__dict__.update(dict)
        return obj


@verify_webhook_secret_sha1(header='sha1=')
@required_params(ARG_PAYLOAD)
@exception_handler
def github_properties_populate(*args, **kwargs):
    """Populate the property configs from the github webhook payload.
    [TODO] Implement the function to address 'remove' action.
    """

    payload = kwargs.get(ARG_PAYLOAD).json
    payload = json.loads(json.dumps(payload), object_hook=Generic.from_dict)
    configs = []
    for commit in payload.commits:
        # only populate the configs attaching 'populate' in the commit message
        if commit.message.startswith('populate'):
            configs += commit.modified
            configs += commit.added

    # sort out config in dict
    config_mapping = defaultdict(list)
    for cfg in configs:
        matched = re.search(r'^Data/(\w+)/.+', cfg)
        if matched:
            config_mapping[matched[1]].append(os.path.basename(os.path.splitext(cfg)[0]))

    # only update when having valid configs to be updated
    for key, value in config_mapping.items():
        cmd_payload = {
            'user': 'U014SE6JH7Y',
            'command': ' '.join([ARG_POPULATE, f'{ARG_CLASS}={key}', *value]),
            'channel': 'T015BP2HUU9#C015DF11US0',
            'response': 'SlackChannel'
        }

        create_command_payload(args=cmd_payload)
